#include <ctype.h>	// isspace
#include <fcntl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>	// strchr
#include <sys/ioctl.h>
#include <unistd.h>

#include "Si5345-Registers.h"

static int i2c_reg_write8(int dev_file, unsigned char dev_addr, unsigned char reg_addr, unsigned char reg_data)
{
    unsigned char outbuf[2];
    struct i2c_rdwr_ioctl_data packets;
    struct i2c_msg messages[1];

    messages[0].addr  = dev_addr;
    messages[0].flags = 0;
    messages[0].len   = sizeof(outbuf);
    messages[0].buf   = outbuf;
    /* The first byte indicates which register we'll write */
    outbuf[0] = reg_addr;
    outbuf[1] = reg_data;
    /* Transfer the i2c packets to the kernel and verify it worked */
    packets.msgs  = messages;
    packets.nmsgs = 1;
    if (ioctl(dev_file, I2C_RDWR, &packets) < 0) {
        perror("Unable to send data");
        return 1;
    }

    return 0;
}

static int _last_page = -1;

static int i2c_reg_write16(int dev_file, unsigned char dev_addr, unsigned short reg_addr, unsigned char reg_data)
{
    const unsigned char page = reg_addr >> 8;
    if (_last_page<0 || _last_page!=page) {
        const int r = i2c_reg_write8(dev_file, dev_addr, 0x01, page);
        if (r != 0) {
            return r;
        }
        _last_page = page;
    }

    return i2c_reg_write8(dev_file, dev_addr, reg_addr & 0xFFu, reg_data);
}

static unsigned int uint_of(const char* s)
{
    if (s[0]=='0' && (s[1]=='x' || s[1]=='X')) {
        return strtol(s, NULL, 16);
    } else {
        return atoi(s);
    }
}

/// Trim the string both from left and from right.
static char* trim(char* s)
{
    // Trim left.
    while (*s!=0 && isspace(*s)) {
        ++s;
    }

    // Trim right.
    char *space_pos = NULL;
    char *s2;
    for (s2=s; *s2!=0; ++s2) {
        if (isspace(*s2)) {
            if (space_pos != NULL) {
                space_pos = s2;
            }
        } else {
            space_pos = NULL;
        }
    }
    if (space_pos != NULL) {
        *space_pos = 0;
    }
    return s;
}

/// Initialize an I2C with the data read from the initialization file.
static int i2c_init_from_file(int i2c_file, const unsigned int i2c_addr, const char* filename)
{
    FILE *fin = fopen(filename, "r");
    char  s_line[500] = { 0 };
    unsigned int ok_count = 0;

    if (fin == NULL) {
        fprintf(stderr, "Cannot open initialization file %s\n", filename);
        return 1;
    }
    while (!feof(fin)) {
        const char* r_fgets = fgets(s_line, sizeof(s_line), fin);
        if (r_fgets == NULL) {
            break;
        }
        char *left_pos = strchr(s_line, '{');
        if (left_pos == NULL) {
            continue;
        }
        const char *comment_pos = strstr(s_line, "//");
        if (comment_pos!=NULL && comment_pos<left_pos) {
            continue;
        }
        char *right_pos = strchr(left_pos, '}');
        if (right_pos == NULL) {
            continue;
        }
	*right_pos = 0;
        char *comma_pos = strchr(left_pos, ',');
        if (comma_pos == NULL) {
            continue;
        }
	*comma_pos = 0;

        // We should have the data to be written by now.
        const char *s_addr = trim(left_pos+1);
        const unsigned int addr = uint_of(s_addr);
        const char *s_val = trim(comma_pos+1);
        const unsigned int val = uint_of(s_val);

        const int r_write = i2c_reg_write16(i2c_file, i2c_addr, addr, val);
        if (r_write == 0) {
            ++ok_count;
            printf(".");
            fflush(stdout);
        } else {
            printf("\n");
            fprintf(stderr, "Couldn't write value 0x%02X to address 0x%04X.\n", val, addr);
            fprintf(stderr, "There were %u succesful I2C writes so far.\n", ok_count);
            fclose(fin);
            return r_write;
        }
    }

    fclose(fin);

    printf("\n");
    if (ok_count>0) {
        printf("Wrote %u values to the I2C bus.\n", ok_count);
        return 0;
    } else {
        fprintf(stderr, "Error: No values found in the initialization file.\n");
        fprintf(stderr, "Nothing written to the I2C bus.\n");
        return 1;
    }
}

int main(int argc, char *argv[])
{
    int i2c_file;

    if (argc < 3) {
        fprintf(stderr, "Not enough arguments.\n");
        fprintf(stderr, "Usage:\n");
        fprintf(stderr, "%s /dev/i2c-X dev_addr            -- Write built-in SI534x initialization sequence.\n", argv[0]);
        fprintf(stderr, "%s /dev/i2c-X dev_addr addr  data -- Single register write.\n", argv[0]);
        fprintf(stderr, "%s /dev/i2c-X dev_addr i2c_header -- Write initialization sequence from the given file.\n", argv[0]);
        return 1;
    }
    const char* i2c_filename = argv[1];
    const unsigned int dev_addr = uint_of(argv[2]);

    if ((i2c_file = open(i2c_filename, O_RDWR)) < 0) {
        perror("Unable to open i2c control file");
        return 1;
    }

    int r = 0;
    if (argc >= 5) {
        const unsigned short writeonly_addr = uint_of(argv[3]);
        const unsigned char  writeonly_data = uint_of(argv[4]);
        printf("I2C write %s: device address 0x%02X: 0x%02X << 0x%02X\n", i2c_filename, dev_addr, writeonly_addr, writeonly_data);
        r = i2c_reg_write16(i2c_file, dev_addr, writeonly_addr, writeonly_data);
    } else if (argc >= 4) {
        // Is the third argument a filename or a hex number?
        const char *filename = argv[3];
        printf("I2C init %s: device address 0x%02X, initialization file %s\n", i2c_filename, dev_addr, filename);
        r = i2c_init_from_file(i2c_file, dev_addr, filename);
    } else {
        int i;
        printf("I2C init %s: device address 0x%02X, built-in SI534X sequence.\n", i2c_filename, dev_addr);
        for (i=0; i<SI5345_REVB_REG_CONFIG_NUM_REGS; ++i) {
            const unsigned char val = si5345_revb_registers[i].value;
            const unsigned short addr = si5345_revb_registers[i].address;
            r = i2c_reg_write16(i2c_file, dev_addr, addr, val);
            if (r != 0) {
                fprintf(stderr, "Couldn't write value 0x%02X to address 0x%04X.\n", val, addr);
                fprintf(stderr, "There were %d succesful I2C writes so far.\n", i);
                break;
            }
        }
    }

    close(i2c_file);
    return r;
}

